#include "scene.hpp"
#include <pew/tools/save.hpp>

#if defined (__EMSCRIPTEN__)
#include <emscripten.h>
#endif

namespace pew {

    Scene::Scene(const Extent2D& extent) :
        extent(extent), renderPass(collection, extent), window(extent) {
        window.setShouldCloseCallback([&] { this->onShouldClose(); });
        window.setKeyInputCallback(
            [&](SDL_KeyboardEvent event, bool down) { this->onKeyInput(event, down); });
        window.addBuffer(&renderPass.getBuffers().color);
        std::stringstream str;
        str << "Rendering at " << extent.width << "x" << extent.height;
        window.setTitle(str.str());
    }

#if defined(PEW_USE_THREADS)
    void
    Scene::render(uint32_t threadCount, bool detectEdges, bool showWindow) {
        if (showWindow) window.startAndFork();

        if (detectEdges) {
            renderPass.renderHitStage(threadCount)->join();
            hitBufferAsColor = renderPass.hitBufferAsColor();
            window.addBuffer(&hitBufferAsColor);
        }

        job = renderPass.renderColorStage(threadCount, detectEdges, true);
    }
#else
    void
    Scene::render(uint32_t threadCount, bool detectEdges, bool showWindow) {
        if (detectEdges) {
            renderPass.renderHitStage();
            hitBufferAsColor = renderPass.hitBufferAsColor();
            window.addBuffer(&hitBufferAsColor);
        }

        renderPass.setPixelUpdateFunction([&] {
            window.handleInput();
            window.drawFrame();
            return window.getShouldClose();
        });
        
        #if defined(__EMSCRIPTEN__)
        renderPass.renderColorStage(detectEdges, true);
        emscripten_set_main_loop_arg([](void* p){((RenderPass*)p)->renderColorStage(true, true);}, &renderPass, -1, 1);
        #else
        renderPass.renderColorStage(detectEdges, true);
        #endif
    }
#endif

#if defined(PEW_USE_THREADS)
    void
    Scene::waitIdle() {
        if (job) job->join();
        window.join();
    }
#endif

    void
    Scene::onShouldClose() {
        window.closeWindow();
#if defined(PEW_USE_THREADS)
        if (job) job->stop();
#endif
    }

#if defined(PEW_USE_THREADS)
    void
    Scene::onKeyInput(SDL_KeyboardEvent event, bool down) {
        if (down) {
            switch (event.keysym.sym) {
                case SDLK_u: window.setIndex(0); break;
                case SDLK_i:
                    if (window.getBuffers().size() > 1) window.setIndex(1);
                    break;

                case SDLK_SPACE: {
                    if (job) { job->paused() ? job->resume() : job->pause(); }
                    break;
                }
                case SDLK_s: {
                    savePng(*window.getCurrentBuffer(), extent, "shot.png");
                    break;
                }
                case SDLK_q: window.closeWindow();
                default: break;
            }
        }
    }
#else
    void
    Scene::onKeyInput(SDL_KeyboardEvent event, bool down) {
        if (down) {
            switch (event.keysym.sym) {
                case SDLK_u: window.setIndex(0); break;
                case SDLK_i:
                    if (window.getBuffers().size() > 1) window.setIndex(1);
                    break;
                case SDLK_s: {
                    savePng(*window.getCurrentBuffer(), extent, "shot.png");
                    break;
                }
                default: break;
            }
        }
    }
#endif

} // namespace pew